/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	The linkable stuff.  We may specialize some templates here.

****************************************************************************/

#include "all.h"

/////////////////////////////////////////////////////////////////////////////
//  Read in the data for a bitmap from a file.
void Bitmap::Import(const FString & name) {
	Destroy();
	g_monitor.ImportFile(name, *this);
	MakeDirty();
}

/////////////////////////////////////////////////////////////////////////////
//  Export my data to a bmp file.
void Bitmap::Export(const FString & name) {
	g_monitor.ExportFile(name, *this);
}

/////////////////////////////////////////////////////////////////////////////
//  Extract from another bitmap.
void Bitmap::ExtractFrom(Map < Color > & m,
						 u16 x,
						 u16 y,
						 u16 w,
						 u16 h) {
	u16 xx, yy;

	Allocate(w, h, __FILE__, __LINE__);
	m.Open(READ);
	Open(WRITE);
	for (yy=0; yy<h; ++yy) {
		for (xx=0; xx<w; ++xx) {
			(*this)[yy][xx] = m[y + yy][x + xx];
		}
	}
	Close();
	m.Close();
	MakeDirty();
}

/////////////////////////////////////////////////////////////////////////////
//  Give me black and white data and I will give you an RLE.  In the future,
// we will want to account for greyscale for antialiasing.
void Bitmap::MakeRLE(color key) {
	List < RLE > rle;
	List < Color > colors;
	u16 y=0;
	u16 x=0;
	u16 countR = 0;
	s16 countX = 0;
	u16 countY = 0;

	SafeDelete(m_rle);
	Open(READ);
	do {
		RLE & cur = rle.Grow();

		// Find out our count.
		while (y < GetHeight()) {
			while (x < GetWidth()) {
				if ((*this)[y][x] != key) {
					goto EndJump;
				}
				++countX;
				++x;
			}
			countX -= GetWidth();
			++countY;
			++y;
			x=0;
		}
		cur.jumpX = countX;
		cur.jumpY = countY;
		cur.run = 0;
		break;

EndJump:
		cur.jumpX = countX;
		cur.jumpY = countY;
		countX = 0;
		countY = 0;
		countR = 0;
		while (x < GetWidth()) {
			if (((*this)[y][x] == key)) {
				goto EndRun;
			}
			colors += (*this)[y][x];
			++countR;
			++x;
		}
		++y;
		x=0;
		cur.run = countR;
		countY = 1;
		countX = -(s16)GetWidth();
		continue;
EndRun:
		cur.run = countR;
		countX = 0;
		countY = 0;
	} while (y < GetHeight());
	Close();

	SafeNew(m_rle, 1, __FILE__, __LINE__);
	m_rle->key = key;
	m_rle->image = rle;
	m_rle->colorData = colors;
}

/////////////////////////////////////////////////////////////////////////////
//  Render a bitmap onto the target.  HERE we may want to make a foreground
// and background color.
void Bitmap::Render(Receptor & r,
					u16 x0,
					u16 y0,
					RenderOptions & ro) {
	r.Open(r.WRITE);
	Open(READ);
	if (ro.blend) {
		ro.blend->Open(READ);
	}
	switch(ro.options) {
		case 0: {
			MType x, y;
		
			for (y=0; y<GetHeight(); ++y) {
				for (x=0; x<GetWidth(); ++x) {
					r[y + y0][x + x0] = (*this)[y][x];
				}
			}
			break;
		}
		case TRANSLUCENT: {
			MType x, y;
			color mask = r.MapColor(0.25, 0.25, 0.25);

			if (ro.blend) {
				for (y=0; y<GetHeight(); ++y) {
					for (x=0; x<GetWidth(); ++x) {
						color c = (((*ro.blend)[y][x] >> 2) & mask) +
								  (((*this)[y][x] >> 2) & mask) * 3;
						r[y + y0][x + x0] = c;
					}
				}
			} else {
				for (y=0; y<GetHeight(); ++y) {
					for (x=0; x<GetWidth(); ++x) {
						color c = ((r[y + y0][x + x0] >> 2) & mask) +
								  (((*this)[y][x] >> 2) & mask) * 3;
						r[y + y0][x + x0] = c;
					}
				}
			}
			break;
		}
		case CUTOUT: {
			MType l, n, o=0;

			BuildRLE(ro.key);
			for (l=0; l<m_rle->image.Count(); ++l) {
				x0 += m_rle->image[l].jumpX;
				y0 += m_rle->image[l].jumpY;
				for (n=0; n<m_rle->image[l].run; ++n) {
					r[y0][x0++] = m_rle->colorData[o++];
				}
			}
			break;
		}
		case TRANSLUCENT | CUTOUT: {
			MType l, n, o=0;
			color mask = r.MapColor(0.25, 0.25, 0.25);

			BuildRLE(ro.key);
			if (ro.blend) {
				for (l=0; l<m_rle->image.Count(); ++l) {
					x0 += m_rle->image[l].jumpX;
					y0 += m_rle->image[l].jumpY;
					for (n=0; n<m_rle->image[l].run; ++n) {
						// HERE might want to pre-compute the RLEs in this format.
						// Really, we need to effect a number of different kinds
						// of rendering strategies based on perceived frequency
						// of changes vs. renderings.
						color c = (((*ro.blend)[y0][x0] >> 2) & mask) +
								  (((m_rle->colorData[o++] >> 2) & mask) * 3);
						r[y0][x0] = c;
					}
				}
			} else {
				for (l=0; l<m_rle->image.Count(); ++l) {
					x0 += m_rle->image[l].jumpX;
					y0 += m_rle->image[l].jumpY;
					for (n=0; n<m_rle->image[l].run; ++n) {
						// HERE might want to pre-compute the RLEs in this format.
						// Really, we need to effect a number of different kinds
						// of rendering strategies based on perceived frequency
						// of changes vs. renderings.
						color c = ((r[y0][x0] >> 2) & mask) +
								  (((m_rle->colorData[o++] >> 2) & mask) * 3);
						r[y0][x0] = c;
					}
				}
			}
			break;
		}		
		case COLORIZE: {
			MType x, y;
		
			for (y=0; y<GetHeight(); ++y) {
				for (x=0; x<GetWidth(); ++x) {
					if ((*this)[y][x] == ro.key) {
						r[y + y0][x + x0] = ro.bg;
					} else {
						r[y + y0][x + x0] = ro.colorize;
					}
				}
			}
			break;
		}
		case COLORIZE | CUTOUT: {
			MType l, n, o=0;

			BuildRLE(ro.key);
			for (l=0; l<m_rle->image.Count(); ++l) {
				x0 += m_rle->image[l].jumpX;
				y0 += m_rle->image[l].jumpY;
				for (n=0; n<m_rle->image[l].run; ++n) {
					r[y0][x0++] = ro.colorize;
				}
			}
			break;
		}
		case COLORIZE | TRANSLUCENT : {
			MType x, y;
			color mask = r.MapColor(0.25, 0.25, 0.25);
			color colorize = ((ro.colorize >> 2) & mask) * 3;
			color bg = ((ro.bg >> 2) & mask) * 3;

			if (ro.blend) {
				for (y=0; y<GetHeight(); ++y) {
					for (x=0; x<GetWidth(); ++x) {
						color c;

						if ((*this)[y][x] == ro.key) {
							c = (((*ro.blend)[y][x] >> 2) & mask) +
								bg;
						} else {
							c = (((*ro.blend)[y][x] >> 2) & mask) +
								colorize;
						}
						r[y + y0][x + x0] = c;
					}
				}
			} else {
				for (y=0; y<GetHeight(); ++y) {
					for (x=0; x<GetWidth(); ++x) {
						color c = ((r[y + y0][x + x0] >> 2) & mask) +
								  colorize;
						r[y + y0][x + x0] = c;
					}
				}
			}
			break;
		}
		case COLORIZE | TRANSLUCENT | CUTOUT : {
			MType l, n, o=0;
			color mask = r.MapColor(0.25, 0.25, 0.25);
			color colorize = ((ro.colorize >> 2) & mask) * 3;
			BuildRLE(ro.key);
			if (ro.blend) {
				for (l=0; l<m_rle->image.Count(); ++l) {
					x0 += m_rle->image[l].jumpX;
					y0 += m_rle->image[l].jumpY;
					for (n=0; n<m_rle->image[l].run; ++n) {
						color c = (((*ro.blend)[y0][x0] >> 2) & mask) + colorize;
						r[y0][x0++] = c;
					}
				}
			} else {
				for (l=0; l<m_rle->image.Count(); ++l) {
					x0 += m_rle->image[l].jumpX;
					y0 += m_rle->image[l].jumpY;
					for (n=0; n<m_rle->image[l].run; ++n) {
						color c = ((r[y0][x0] >> 2) & mask) + colorize;
						r[y0][x0++] = c;
					}
				}
			}
			break;
		}
	}
	if (ro.blend) {
		ro.blend->Close();
	}
	Close();
	r.Close();
}

/////////////////////////////////////////////////////////////////////////////
//  Map a bitmap's color:  use the monitor mapping.
Color Bitmap::MapColor(real r, real g, real b) const {
	return g_monitor.MapColor(r, g, b);
}

/////////////////////////////////////////////////////////////////////////////
//  Clear the indicated portion of the target.
void Bitmap::Clear(Color u, Rect * p) {
	u16 x, y, w, h;

	if (p) {
		x = p->x;
		y = p->y;
		w = p->w;
		h = p->h;
	} else {
		x = 0;
		y = 0;
		w = GetWidth();
		h = GetHeight();
	}

	Open(WRITE);
	while(h--) {
		u16 ww = w;

		while (ww--) {
			(*this)[y + h][x + ww] = u;
		}
	}
	Close();
}

/////////////////////////////////////////////////////////////////////////////
//  When we render an overlay, if we don't have hardware overlay support, we
// queue the rendering 'til after everything else (including transparencies).
void Overlay::Render(Receptor & r,
					 u16 x,
					 u16 y,
					 RenderOptions & ro) {
	r.QueueOverlay(this, x, y, ro);
}

/////////////////////////////////////////////////////////////////////////////
//  Receptors have a predetermined sequence for the stuff it renders.  This
// will do the actual rendering, presumably at the end of the frame.  We also
// save the underlying graphic data.
void Overlay::ActuallyDraw(Receptor & r,
						   u16 x,
						   u16 y,
						   RenderOptions & ro) {
	Save & s = m_saveData[r.GetCurrentBuffer()];
	Bitmap & b = s.bitmap;

	BuildRLE(ro.key);
	if (ro.options & CUTOUT) {
		MType l, n, o=0;
		u16 xx=x;
		u16 yy=y;

		if (!b.m_rle) {
			SafeNew(b.m_rle, 1, __FILE__, __LINE__);
			b.m_rle->colorData.Make(m_rle->colorData.Count(), __FILE__, __LINE__);
			b.m_rle->image = m_rle->image;
		}
		b.m_rle->key = ro.key;

		r.Open(r.READ);
		for (l=0; l<b.m_rle->image.Count(); ++l) {
			xx += b.m_rle->image[l].jumpX;
			yy += b.m_rle->image[l].jumpY;
			for (n=0; n<m_rle->image[l].run; ++n) {
				b.m_rle->colorData[o++] = r[yy][xx++];
			}
		}
		r.Close();
		s.options = CUTOUT;
		s.key = ro.key;
	} else {
		b.ExtractFrom(r, x, y, GetWidth(), GetHeight());
		s.options = 0;
	}
	Bitmap::Render(r, x, y, ro);
}

/////////////////////////////////////////////////////////////////////////////
//  Undo the damage done by our last appearance on the target receptor.
void Overlay::Erase(Receptor & r, u16 x, u16 y) {
	Save & s = m_saveData[r.GetCurrentBuffer()];
	RenderOptions ro;

	ro.options = s.options;
	ro.key = s.key;
	s.bitmap.Render(r, x, y, ro);
}

/////////////////////////////////////////////////////////////////////////////
//  Dirty this bitmap.
void Bitmap::MakeDirty(void) {
	SafeDelete(m_rle);
}

/////////////////////////////////////////////////////////////////////////////
//  When we dirty an overlay, we have to destroy all of the dependencies.
void Overlay::MakeDirty(void) {
	MType m;

	Bitmap::MakeDirty();
	for (m=0; m<m_saveData.Count(); ++m) {
		SafeDelete(m_saveData[m].bitmap.m_rle);
	}
}
