/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	The non-inline implementation of istrings.

****************************************************************************/

#include "all.h"

/////////////////////////////////////////////////////////////////////////////
//  Make an IString out of an ascii string.
IString::IString(const char * s) {
	if (s) {
		MType m = strlen(s);

		m_nextAlloc = 1;
		while (m_nextAlloc < m) {
			m_nextAlloc += m_nextAlloc;
		}
		m_numberOfElements = m;
		SafeNew(m_elements, m_nextAlloc, __FILE__, __LINE__);
		while (m--) {
			m_elements[m] = (wchar_t)s[m];
		}
	} else {
		m_nextAlloc = 1;
		m_numberOfElements = 0;
		m_elements = NULL;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Make an IString out of a wide character.
IString::IString(const wchar_t s) {
	m_nextAlloc = 1;
	m_numberOfElements = 1;
	SafeNew(m_elements, (MType)1, __FILE__, __LINE__);
	m_elements[0] = s;
}

/////////////////////////////////////////////////////////////////////////////
//  Copy a wchar_t to an IString.
IString & IString::operator =(const wchar_t s) {
	SafeDelete(m_elements);
	m_nextAlloc = 1;
	m_numberOfElements = 1;
	SafeNew(m_elements, (MType)1, __FILE__, __LINE__);
	m_elements[0] = s;
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Copy an ascii string into an IString.
IString & IString::operator =(const char * s) {
	MType m = strlen(s);

	SafeDelete(m_elements);
	m_nextAlloc = 1;
	while (m_nextAlloc < m) {
		m_nextAlloc += m_nextAlloc;
	}
	m_numberOfElements = m;
	SafeNew(m_elements, m_nextAlloc, __FILE__, __LINE__);
	while (m--) {
		m_elements[m] = (wchar_t)s[m];
	}
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Copy an FString to an IString.
IString & IString::operator =(const FString & f) {
	MType m = f.Length();

	SafeDelete(m_elements);
	m_nextAlloc = 1;
	while (m_nextAlloc < m) {
		m_nextAlloc += m_nextAlloc;
	}
	m_numberOfElements = m;
	SafeNew(m_elements, m_nextAlloc, __FILE__, __LINE__);
	while (m--) {
		m_elements[m] = f[m];
	}
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Compare one IString to another.
Bool IString::operator <(const IString & i) const {
	MType m, u;

	u = Min(Length(), i.Length());

	for (m=0; m<u; ++m) {
		wchar_t a = (*this)[m];
		wchar_t b = i[m];

		if (a < b) {
			return TRUE;
		} else if (a > b) {
			return FALSE;
		}
	}
	if (u != i.Length()) {
		return TRUE;
	} else {
		return FALSE;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Write this IString to debug output.  
void IString::DebugOut(void) const {
	#if defined(WIN32)
		char * buf = (char *)alloca(m_numberOfElements + 1);
	
		if (buf) {
			MType m;

			for (m=0; m<m_numberOfElements; ++m) {
				buf[m] = (char)m_elements[m];
			}
			buf[m] = '\0';
			OutputDebugString(buf);
		}
	#endif
}

/////////////////////////////////////////////////////////////////////////////
//  Write this FString to debug output.
void FString::DebugOut(void) const {
	#if defined(WIN32)
		char buf[FSTRING_SIZE + 1];
		MType m;

		for (m=0; m<m_cursor; ++m) {
			buf[m] = (char)m_elements[m];
		}
		buf[m] = '\0';
		OutputDebugString(buf);
	#endif
}

/////////////////////////////////////////////////////////////////////////////
//  Compare one FString to another.
Bool FString::operator <(const FString & i) const {
	MType m, u;

	u = Min(Length(), i.Length());

	for (m=0; m<u; ++m) {
		wchar_t a = (*this)[m];
		wchar_t b = i[m];

		if (a < b) {
			return TRUE;
		} else if (a > b) {
			return FALSE;
		}
	}
	if (u != i.Length()) {
		return TRUE;
	} else {
		return FALSE;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  For now.  In the future, we will have to modify this to actually consult
// the registry, ini files, environment, and command line.
inline const char * GetENVVar(const char * var) {
	return "english.csv";
}

/////////////////////////////////////////////////////////////////////////////
//  This class manages the translation of text between english and other
// languages.
class Translator {
	public:
		#if !defined(RELEASE_BUILD)
			friend void WriteTable(void);
		#endif

		static const char * Lookup(const char * s);
	protected:
	private:
		static MType GetHash(const char * s);

		struct Entry {
			Entry * next;
			char * untranslated;
			char * translated;
		};
		static Entry * s_entries[256];
		static char * s_language;
};

char * Translator::s_language;
Translator::Entry * Translator::s_entries[256];

/////////////////////////////////////////////////////////////////////////////
//  Make a hash out of the indicated string.
MType Translator::GetHash(const char * s) {
	u8 retval=0;

	while (*s) {
		retval += (u8)*s++;
	}
	return retval & 0xff;
}

/////////////////////////////////////////////////////////////////////////////
//  Look up the indicated string in the string table.
const char * Translator::Lookup(const char * s) {
	Entry * e;
	MType hash;

	if (!s_language) {
		FILE * f;
		char buf[2048];

		s_language = strdup(GetENVVar("language"));
		if (!(f = fopen(s_language, "r"))) {
			Warn("Unable to open %s for reading.\n", s_language);
		} else {
			while (!feof(f)) {
				if (fgets(buf, 2048, f)) {
					char * toks[3];

					toks[0] = strtok(buf, "\"");
					toks[1] = strtok(NULL, "\"");
					toks[1] = strtok(NULL, "\"");
					toks[2] = strtok(NULL, "\"");
					if (toks[0] && toks[1]) {
						SafeNew(e, (MType)1, __FILE__, __LINE__);
						hash = GetHash(toks[0]);

						e->next = s_entries[hash];
						s_entries[hash] = e;
						e->untranslated = strdup(toks[0]);
						e->translated = strdup(toks[1]);
					}
				}
			}
		}
	}
	hash = GetHash(s);
	e = s_entries[hash];
	while(e) {
		if (!strcmp(e->untranslated, s)) {
			break;
		}
		e=e->next;
	}
	if (!e) {
		SafeNew(e, (MType)1, __FILE__, __LINE__);

		e->next = s_entries[hash];
		s_entries[hash] = e;
		e->untranslated = strdup(s);
		e->translated = strdup(s);
	}
	return e->translated;
}

#if !defined(RELEASE_BUILD)
	/////////////////////////////////////////////////////////////////////////////
	//  A shutdown routine that rewrites the file.  Only in development builds.
	static void WriteTable(void) {
		MType m;
		FILE * f;
	
		if (!(f = fopen(Translator::s_language, "w"))) {
			Warn("Unable to open %s for writing.\n", Translator::s_language);
		} else {
			for (m=0; m<256; ++m) {
				Translator::Entry * e;
		
				for (e=Translator::s_entries[m]; e;) {
					Translator::Entry * next = e->next;

					fprintf(f,
							"\"%s\", \"%s\"\n",
							e->untranslated,
							e->translated);
					free(e->translated);
					free(e->untranslated);
					SafeDelete(e);
					e = next;
				}
				Translator::s_entries[m] = NULL;
			}
			free(Translator::s_language);
			Translator::s_language = NULL;
		}
	}

	static DeInitializer s_writeLanguage(WriteTable);

#endif

/////////////////////////////////////////////////////////////////////////////
//  A TString is a string that is automatically translated into a different
// language.
TString::TString(const char * s) : IString(Translator::Lookup(s)) {
}

/////////////////////////////////////////////////////////////////////////////
//  Are these IStrings different.
Bool IString::operator !=(const IString & i) const {
	MType m;

	if (Length() != i.Length()) {
		return TRUE;
	}
	for (m=0; m<Length(); ++m) {
		if ((*this)[m] != i[m])
			return TRUE;
	}
	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////
//  Are these FStrings different.
Bool FString::operator !=(const FString & i) const {
	MType m;

	if (Length() != i.Length()) {
		return TRUE;
	}
	for (m=0; m<Length(); ++m) {
		if ((*this)[m] != i[m])
			return TRUE;
	}
	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////
//  Truncate the string.
void FString::Chop(MType nChars) {
	if (m_cursor >= nChars) {
		m_cursor -= nChars;
	} else {
		m_cursor = 0;
	}
}
