/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	The inlines and template members of jmath.

****************************************************************************/

/////////////////////////////////////////////////////////////////////////////
//  Cast to array of floats.
template < class Elem,
		   MType t_dim >
inline
Elem *
MathVec < Elem, t_dim >::CastToFloatPtr(void) {
	return m_elements;
}

/////////////////////////////////////////////////////////////////////////////
//  Subtract a vector - non-member function.
template < class Elem,
		   MType t_dim >
inline
MathVec < Elem, t_dim > operator -(const MathVec < Elem, t_dim > & v,
								   const MathVec < Elem, t_dim > & w) {
	MathVec < Elem, t_dim > retval;
	MType m;

	for (m=0; m<t_dim; ++m) {
		retval[m] = v[m] - w[m];
	}
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Multiply a vector by a scalar.
template < class Elem, MType t_dim >
inline
MathVec < Elem, t_dim > operator *(const MathVec < Elem, t_dim > & v,
								   Elem r) {
	MathVec < Elem, t_dim > retval;
	MType m;

	for (m=0; m<t_dim; ++m) {
		retval[m] = v[m] * r;
	}
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Do a dot product.
template < class Elem, MType t_dim >
Elem operator *(const MathVec < Elem, t_dim > & v,
				const MathVec < Elem, t_dim > & w) {
	MType m;
	Elem retval = (Elem)0;

	for (m=0; m<t_dim; ++m) {
		retval += v[m] * w[m];
	}
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Initialize a vector to some constant value.
template < class Elem,
		   MType t_dim >
inline
MathVec < Elem, t_dim > &
MathVec < Elem, t_dim >::operator =(Elem e) {
	MType m;

	for (m=0; m<t_dim; ++m)
		m_elements[m] = e;
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Add to a vector.
template < class Elem,
		   MType t_dim >
inline
MathVec < Elem, t_dim > &
MathVec < Elem, t_dim >::operator +=(const MathVec < Elem, t_dim > & w) {
	MType u;

	for (u=0; u<t_dim; ++u) {
		m_elements[u] += w[u];
	}
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Negate this vector.
template < class Elem, MType t_dim >
void
MathVec < Elem, t_dim >::Negate(void) {
	m_elements[0] = -m_elements[0];
	m_elements[1] = -m_elements[1];
	m_elements[2] = -m_elements[2];
}

/////////////////////////////////////////////////////////////////////////////
//  Make a unit vector out of this puppy.  Return the former magnitude of the
// vector.  HEREHEREHERE what kind of calculator do we wish to use?
template < class Elem, MType t_dim >
Elem
MathVec < Elem, t_dim >::Unit(void) {
	Elem retval=0;
	Elem rsq;
	MType u;

	for (u=0; u<t_dim; ++u) {
		retval += m_elements[u] * m_elements[u];
	}
	if (retval > 1.0e-37) {
		rsq = Recipsqrt(retval);
		for (u=0; u<t_dim; ++u) {
			m_elements[u] *= rsq;
		}
		return rsq;
	} else {
		return 0;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Let's invert the indicated matrix.
template < class Elem, MType t_dim >
Bool
SqrMatrix < Elem, t_dim >::Invert(void) {
	if (t_dim) {
		if (t_dim == 1) {
			if (Abs(m_elements[0][0]) > Epsilon(m_elements[0][0])) {
				m_elements[0][0] = Reciprocoal(m_elements[0][0]);
				return TRUE;
			} else {
				return FALSE;
			}
			
		} else if (t_dim == 2) {
			Elem det = m_elements[0][0] * m_elements[1][1] -
					   m_elements[0][1] * m_elements[1][0];
			Elem a = m_elements[0][0];

			if (det > Epsilon(det)) {
				det = Reciprocoal(det);
				m_elements[0][0] = det * m_elements[1][1];
				m_elements[0][1] *= -det;
				m_elements[1][0] *= -det;
				m_elements[1][1] = det * a;
				return TRUE;
			} else {
				return FALSE;
			}
		} else {
			return FALSE;		// NYI.  Inverse, if very large, Seidel.
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Our low-precision calculator.  This calculator does all operations with
// speed alone in mind, precision suffers for this.
inline u16 Calculator < JM_LOW_PRECISION >::RadToJDeg(real r) {
	// HEREHEREHERE not yet optimized.
	u16 retval = (u16)((r * JM_DEGREES) / (2 * JM_PI));

	retval &= (JM_DEGREES - 1);
	return retval;
}

/////////////////////////////////////////////////////////////////////////////
//  Compute the cosine, given optimized degrees.
inline real Calculator < JM_LOW_PRECISION >::JCos(u16 degs) {
	return g_cosTable[(degs >> 6) & (JM_DEGREES - 1)];
}

/////////////////////////////////////////////////////////////////////////////
//  Compute the sine, given optimized degrees.
inline real Calculator < JM_LOW_PRECISION >::JSin(u16 degs) {
	return JCos(degs - JM_DEGREES / 4);
}

/////////////////////////////////////////////////////////////////////////////
//  Compute the cosine, given non-optimized degrees.
inline real Calculator < JM_LOW_PRECISION >::Cos(real degs) {
	return JCos(RadToJDeg(degs));
}

/////////////////////////////////////////////////////////////////////////////
//  Compute the sine, given non-optimized degress.
inline real Calculator < JM_LOW_PRECISION >::Sin(real degs) {
	return JSin(RadToJDeg(degs));
}

/////////////////////////////////////////////////////////////////////////////
//  Compute the reciprocoal.  HEREHEREHERE not yet optimized.  LCalc will do
// a lookup, MCalc will do one N-R, LCalc will do as many as needed.
inline real Calculator < JM_LOW_PRECISION >::Reciprocoal(real r) {
	return (real)(1.0 / r);
}

/////////////////////////////////////////////////////////////////////////////
//  Our medium-precision calculator.  A compromise between performance and
// quality.
inline u16 Calculator < JM_MED_PRECISION >::RadToJDeg(real r) {
	return LCalculator.RadToJDeg(r);
}

/////////////////////////////////////////////////////////////////////////////
//  Optimized cosine.
inline real Calculator < JM_MED_PRECISION >::JCos(u16 degs) {
	real r = (degs & 63) * (1.0 / 64);
	real c1, c2;

	degs = (degs >> 6);
	c1 = g_cosTable[degs & 1023];
	c2 = g_cosTable[(degs + 1) & 1023];

	return (c2 * r) + (c1 * (1-r));
}

/////////////////////////////////////////////////////////////////////////////
//  Optimized sine.
inline real Calculator < JM_MED_PRECISION >::JSin(u16 degs) {
	return JCos(degs - JM_DEGREES / 4);
}

/////////////////////////////////////////////////////////////////////////////
//  Normal cosine.
inline real Calculator < JM_MED_PRECISION >::Cos(real degs) {
	return JCos(RadToJDeg(degs));
}

/////////////////////////////////////////////////////////////////////////////
//  Normal sine.
inline real Calculator < JM_MED_PRECISION >::Sin(real degs) {
	return JSin(RadToJDeg(degs));
}

/////////////////////////////////////////////////////////////////////////////
//  Compute the reciprocoal.  HEREHEREHERE not yet optimized.  LCalc will do
// a lookup, MCalc will do one N-R, LCalc will do as many as needed.
inline real Calculator < JM_MED_PRECISION >::Reciprocoal(real r) {
	return (real)(1.0 / r);
}


// HEREHEREHERE have to implement the high-precision calculator.  For this,
// we still favor performance where we can, but we want precise results.  We
// inline only if it's worthwhile.
