/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	Okay, these are our substitutes for pointers, arrays, etc.  The idea is
to automate garbage collection, implement bounds-checking, etc.  Also, we
can plug in our own memory allocation function.

****************************************************************************/

#ifndef PTR_H
	#define PTR_H

	// Why are these here?  They're needed for the forward friend declaration
	// below.
	typedef u8 (*ThreadEntryPoint)(void * arg);
	typedef MType ThreadId;
	typedef u16 ThreadStatus;

	/////////////////////////////////////////////////////////////////////////
	//  Our pointers will use the mark-and-sweep algorithm, by way of a
	// worker thread.
	template < class O >
	class Ptr {
		friend u8 ThreadLaunchThread(ThreadEntryPoint,
									 Ptr < O >,
									 ThreadId &,
									 ThreadStatus &,
									 u8,
									 MType);
		public:
			typedef O Element;

			Ptr(void);
			Ptr(O * o);

			#if !defined(DOS)
				template < class P >
				Ptr(Ptr < P > & p) { AddPointer(); m_object = (O *)&(*p); }
			#endif
			Ptr(Ptr & p) { AddPointer(); m_object = p.m_object; }

			~Ptr(void);

			operator Bool(void) const;

			Ptr & operator =(const Ptr & p);
			Ptr & operator =(O * o);

			O * operator ->(void);
			const O & operator *(void) const;
			O & operator *(void);

			Bool operator ==(const Ptr < O > & p) const;

			void Construct(const char * f, int l);

			void * CastToVoidPointer(void);
		protected:
		private:
			void AddPointer(void);
			void RemovePointer(void);
			static void AddObject(void * o, void (*DestroyFunc)(void *));

			O * m_object;
			MType m_positionInList;

			static void ** s_allObjects;
			static void ** s_allPointers;
	};

	/////////////////////////////////////////////////////////////////////////
	//	Variable-length lists.  HEREHEREHERE we have to put the data
	// allocated by lists and arrays in the same garbage-collection system
	// that we established before.  We may want to demand-copy it as well.
	template < class O >
	class List {
		public:
			List(void);
			List(const List & l);
			~List(void);

			List & operator =(const List & l);

			O & Grow(void);
			List & operator <<=(const O & e);		// Insert if not there.
			List & operator |=(const O & e);		// Insert or replace.
			List & operator +=(const O & e);		// Insert.
			List & operator +=(const List & l);
			List & operator -=(const O & e);
			void RemoveWithoutSavingOrder(const O & e);
			O & operator [](MType m);
			const O & operator [](MType m) const;
			MType Count(void) const;
			void Destroy(void);

			// Remove terminal items.
			void Chop(void);

			// Reverse sequence.
			void Flip(void);
		protected:
			void Copy(const List & l);

			MType m_numberOfElements;
			MType m_nextAlloc;
			O * m_elements;
		private:
	};

	/////////////////////////////////////////////////////////////////////////
	//	Lists whose length is determined at construction time.
	template < class O >
	class Array {
		public:
			Array(void);
			Array(const List < O > & l);
			Array(const Array & a);
			~Array(void);

			Array & operator =(const List < O > & l);
			Array & operator =(const Array < O > & a);

			void Make(MType size, const char * f, int l);
			void Destroy(void);
			O & operator [](MType m);
			const O & operator [](MType m) const;
			MType Count(void) const;
			void * CastToVoidPointer(void);
		protected:
		private:
			O * m_elements;
			MType m_size;
	};

	/////////////////////////////////////////////////////////////////////////
	//	Fixed-length lists.
	template < class O, u8 t_dim >
	class Vec {
		public:
			O & operator [](MType m);
			const O & operator [](MType m) const;
			Vec & operator =(const Vec & v);
			Bool operator ==(const Vec & v) const;
			Bool operator !=(const Vec & v) const;
			MType Count(void) const;
			void Fill(const Array < O > & src,
					  MType start=0);
			void * CastToVoidPointer(void);
		protected:
			O m_elements[t_dim];
		private:
	};

	/////////////////////////////////////////////////////////////////////////
	//	Not for general use - allows us to "tie" an array into another
	// aggregation.
	template < class O >
	class MArray {
		public:
			MArray(MType size, O * o);
			O & operator [](MType m);
			const O & operator [](MType m) const;
			MType Count(void) const;
		protected:
		private:
			O * m_elements;
			MType m_size;
	};

	extern u8 PtrGarbageCollectionThread(void * arg);

	template < class C >
	void SafeNew(C * & c, MType b, const char * f, int l);
	template < class C >
	void SafeDelete(C * & c);
	#if defined(TRACK_MEMORY)
		extern void MemoryDump(void);
	#endif

	#include "ptr.ipp"

#endif
