/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	Templates & inlines for our pointer classes.  A danger we have to worry
about is having a live pointer whose object is dead, but I think this
situation has been avoided.

****************************************************************************/

// This template function "builds" a destructor for every object.
template < class F >
class PtrDestroyDummy {
	public:
		static void Destroy(void * v);
	protected:
	private:
};

template < class F >
void
PtrDestroyDummy < F >::Destroy(void * v) {
	F * f = (F *)v;

	f->~F();
}

extern MType PtrAddPointer(void * p);
extern void PtrRemovePointer(MType which);
extern void PtrAddObject(void * o, void (*DestroyFunc)(void *));

/////////////////////////////////////////////////////////////////////////////
//  These functions will manage the list of objects and pointers that are
// presently alive.
template < class O >
inline
void
Ptr < O >::AddPointer(void) {
	m_positionInList = PtrAddPointer((void *)this);
}


/////////////////////////////////////////////////////////////////////////////
//  Remove the pointer from the global list.
template < class O >
inline
void
Ptr < O >::RemovePointer(void) {
	PtrRemovePointer(m_positionInList);
}

/////////////////////////////////////////////////////////////////////////////
//  Add an object to the global list.  The removal is managed by our garbage
// collection thread.
template < class O >
inline
void
Ptr < O >::AddObject(void * o, void (*DestroyFunc)(void *)) {
	PtrAddObject(o, DestroyFunc);
}

/////////////////////////////////////////////////////////////////////////////
//  Default constructor for a pointer makes it NULL in debug mode only.
template < class O >
inline
Ptr < O >::Ptr(void) {
	AddPointer();
	#if !defined(NDEBUG)
		m_object = NULL;
	#endif
}

/////////////////////////////////////////////////////////////////////////////
//  Make one from a C++ pointer.
template < class O >
inline
Ptr < O >::Ptr(O * o) {
	AddPointer();
	m_object = o;
}

/////////////////////////////////////////////////////////////////////////////
//  Destroy a pointer.
template < class O >
inline
Ptr < O >::~Ptr(void) {
	RemovePointer();
}

/////////////////////////////////////////////////////////////////////////////
//  Check for NULL pointer.
template < class O >
inline
Ptr < O >::operator Bool(void) const {
	return m_object ? TRUE : FALSE;
}

/////////////////////////////////////////////////////////////////////////////
//  Assign one pointer to another.
template < class O >
inline
Ptr < O > &
Ptr < O >::operator =(const Ptr < O > & p) {
	m_object = p.m_object;
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Copy in an object.
template < class O >
inline
Ptr < O > &
Ptr < O >::operator =(O * o) {
	m_object = o;
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Dereference the object that this pointer points to.
template < class O >
inline
O *
Ptr < O >::operator ->(void) {
	CHECK(m_object);
	return m_object;
}

/////////////////////////////////////////////////////////////////////////////
//  Another way to dereference.
template < class O >
inline
const O &
Ptr < O >::operator *(void) const {
	CHECK(m_object);
	return *m_object;
}

/////////////////////////////////////////////////////////////////////////////
//  We have both const and non-const.
template < class O >
inline
O &
Ptr < O >::operator *(void) {
	CHECK(m_object);
	return *m_object;
}

/////////////////////////////////////////////////////////////////////////////
//  Actually construct an object.
template < class O >
#if defined(NDEBUG)
	inline
#endif
void
Ptr < O >::Construct(const char * f, int l) {
	void * v = malloc(sizeof(O));
	
	m_object = new (v) O;
	AddObject(m_object, PtrDestroyDummy < O >::Destroy);
}

/////////////////////////////////////////////////////////////////////////////
//  This is needed by our thread library:  cast this pointer to a void *.
// HERE this violates the safety of the class. . .
template < class O >
void *
Ptr < O >::CastToVoidPointer(void) {
	return (void *)m_object;
}

/////////////////////////////////////////////////////////////////////////////
//  Are these pointers pointing to the same object.
template < class O >
Bool Ptr < O >::operator ==(const Ptr < O > & p) const {
	return m_object == p.m_object;
}

/////////////////////////////////////////////////////////////////////////////
//  Dereference the indicated element in the vec.
template < class O, u8 t_dim >
inline
O & Vec < O, t_dim >::operator [](MType m) {
	CHECK(m < t_dim);
	return m_elements[m];
}

/////////////////////////////////////////////////////////////////////////////
//  Copy one vector to another.
template < class O, u8 t_dim >
inline
Vec < O, t_dim > & Vec < O, t_dim >::operator =(const Vec < O, t_dim > & v) {
	MType m;

	for (m=0; m<Count(); ++m) {
		(*this)[m] = v[m];
	}
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  The const case.
template < class O, u8 t_dim >
inline
const O & Vec < O, t_dim >::operator [](MType m) const {
	CHECK(m < t_dim);
	return m_elements[m];
}

/////////////////////////////////////////////////////////////////////////////
//  Do these two items have identical contents.
template < class O, u8 t_dim >
inline
Bool Vec < O, t_dim >::operator ==(const Vec < O, t_dim > & v) const {
	MType m;

	for (m=0; m<t_dim; ++m) {
		if (v[m] != m_elements[m])
			return FALSE;
	}
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//  Convert this to a void pointer, for where it's necessary.
template < class O, u8 t_dim >
inline
void * Vec < O, t_dim >::CastToVoidPointer(void) {
	return (void *)m_elements;
}

/////////////////////////////////////////////////////////////////////////////
//  Are these two unequal.
template < class O, u8 t_dim >
inline
Bool Vec < O, t_dim >::operator !=(const Vec < O, t_dim > & v) const {
	MType m;

	for (m=0; m<t_dim; ++m) {
		if (v[m] != m_elements[m])
			return TRUE;
	}
	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////
//  Return the size of the array to the caller.
template < class O, u8 t_dim >
inline
MType Vec < O, t_dim >::Count(void) const {
	return t_dim;
}

/////////////////////////////////////////////////////////////////////////////
//  Fill the contents of this array with the contents of another, without
// replacing my memory.
template < class O, u8 t_dim >
void Vec < O, t_dim >::Fill(const Array < O > & src, MType start) {
	MType m;

	CHECK(src.Count() + start <= Count());
	for (m=0; m<src.Count(); ++m) {
		(*this)[start++] = src[m];
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Construct an array.
template < class O >
Array < O >::Array(void) {
	m_elements = NULL;
	m_size = 0;
}

/////////////////////////////////////////////////////////////////////////////
//  Construct an array from a list.
template < class O >
Array < O >::Array(const List < O > & l) {
	MType m;

	m_size = l.Count();
	SafeNew(m_elements, m_size, __FILE__, __LINE__);
	for (m=0; m<m_size; ++m) {
		m_elements[m] = l[m];
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Construct from another array.
template < class O >
Array < O >::Array(const Array & a) {
	MType m;

	m_size = a.Count();
	SafeNew(m_elements, m_size, __FILE__, __LINE__);
	for (m=0; m<m_size; ++m) {
		m_elements[m] = a[m];
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Destroy an array.
template < class O >
Array < O >::~Array(void) {
	Destroy();
}

/////////////////////////////////////////////////////////////////////////////
//  Make an array whose size is determined at run-time.
template < class O >
void Array < O >::Make(MType size, const char * f, int l) {
	if (size != m_size) {
		Destroy();
		SafeNew(m_elements, size, f, l);
		m_size = size;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Clear out the array.
template < class O >
void Array < O >::Destroy(void) {
	SafeDelete(m_elements);
	m_elements = NULL;
	m_size = 0;
}

/////////////////////////////////////////////////////////////////////////////
//  Make an array out of a list.
template < class O >
Array < O > & Array < O >::operator =(const List < O > & l) {
	MType m;

	Destroy();
	m_size = l.Count();
	SafeNew(m_elements, m_size, __FILE__, __LINE__);
	for (m=0; m<m_size; ++m) {
		m_elements[m] = l[m];
	}
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Make an array out of another array.
template < class O >
Array < O > & Array < O >::operator =(const Array < O > & a) {
	MType m;

	Destroy();
	m_size = a.Count();
	SafeNew(m_elements, m_size, __FILE__, __LINE__);
	for (m=0; m<m_size; ++m) {
		m_elements[m] = a[m];
	}
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Dereference l-value.
template < class O >
inline
O & Array < O >::operator [](MType m) {
	CHECK(m < Count());
	return m_elements[m];
}

/////////////////////////////////////////////////////////////////////////////
//  Dereference r-value.
template < class O >
inline
const O & Array < O >::operator [](MType m) const {
	CHECK(m < Count());
	return m_elements[m];
}

/////////////////////////////////////////////////////////////////////////////
//  Get size of array.
template < class O >
inline
MType Array < O >::Count(void) const {
	return m_size;
}

/////////////////////////////////////////////////////////////////////////////
//  In rare occasions, we need this.
template < class O >
inline
void * Array < O >::CastToVoidPointer(void) {
	return (void *)(&m_elements[0]);
}

/////////////////////////////////////////////////////////////////////////////
//  Make a copy of the indicated list.
template < class O >
void
List < O >::Copy(const List & l) {
	MType m;

	m_numberOfElements = l.m_numberOfElements;
	m_nextAlloc = l.m_nextAlloc;
	SafeNew(m_elements, m_nextAlloc, __FILE__, __LINE__);
	for (m=0; m<m_numberOfElements; ++m) {
		m_elements[m] = l.m_elements[m];
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Build a list.
template < class O >
List < O >::List(void) {
	m_numberOfElements = 0;
	m_nextAlloc = 1;
	m_elements = NULL;
}

/////////////////////////////////////////////////////////////////////////////
//  Copy a list.
template < class O >
List < O >::List(const List & l) {
	Copy(l);
}

/////////////////////////////////////////////////////////////////////////////
//  Destroy the list.
template < class O >
List < O >::~List(void) {
	SafeDelete(m_elements);
}

/////////////////////////////////////////////////////////////////////////////
//  Make a copy of the other list.
template < class O >
List < O > &
List < O >::operator =(const List < O > & l) {
	SafeDelete(m_elements);
	Copy(l);
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Grow the list by one.
template < class O >
inline
O & List < O >::Grow(void) {
	if (m_numberOfElements + 1 >= m_nextAlloc) {
		O * rep;
		MType m;

		m_nextAlloc += m_nextAlloc;
		SafeNew(rep, m_nextAlloc, __FILE__, __LINE__);
		for (m=0; m<m_numberOfElements; ++m) {
			rep[m] = m_elements[m];
		}
		SafeDelete(m_elements);
		m_elements = rep;
	}
	return m_elements[m_numberOfElements++];
}

/////////////////////////////////////////////////////////////////////////////
//  Add an element to the list.
template < class O >
inline
List < O > & List < O >::operator +=(const O & e) {
	Grow() = e;
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Add an element to the list given that the element doesn't already exist
// in the list.
template < class O >
inline
List < O > & List < O >::operator <<=(const O & e) {
	MType m;

	for (m=0; m<Count(); ++m) {
		if (m_elements[m] == e) {
			return *this;
		}
	}
	Grow() = e;
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Add an element to the list given that the element doesn't already exist
// in the list.  If it does exist, replace that item with this item (this is
// an updated version of the same item).
template < class O >
inline
List < O > & List < O >::operator |=(const O & e) {
	MType m;

	for (m=0; m<Count(); ++m) {
		if (m_elements[m] == e) {
			m_elements[m] = e;
			return *this;
		}
	}
	Grow() = e;
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Remove an element from the list, if it exists.
template < class O >
inline
List < O > & List < O >::operator -=(const O & e) {
	MType m;

	for (m=0; m<Count(); ++m) {
		if (m_elements[m] == e) {
			MType n;

			for (n=m + 1; n<Count(); ++n) {
				m_elements[n - 1] = m_elements[n];
			}
			--m_numberOfElements;
			break;
		}
	}
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Remove an element with no regard for preserving order.
template < class O >
inline
void List < O >::RemoveWithoutSavingOrder(const O & e) {
	MType m;

	for (m=0; m<Count(); ++m) {
		if (m_elements[m] == e) {
			m_elements[m] = m_elements[--m_numberOfElements];
			break;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  This adds an entire list to another list.
template < class O >
inline
List < O > & List < O >::operator +=(const List < O > & l) {
	MType n = m_numberOfElements + l.m_numberOfElements;

	if (n >= m_nextAlloc) {
		O * rep;
		MType m;

		while (m_nextAlloc < n) {
			m_nextAlloc += m_nextAlloc;
		}
		SafeNew(rep, m_nextAlloc, __FILE__, __LINE__);
		for (m=0; m<m_numberOfElements; ++m) {
			rep[m] = m_elements[m];
		}
		SafeDelete(m_elements);
		m_elements = rep;
	}
	for (n=0; n<l.m_numberOfElements; ++n) {
		m_elements[n + m_numberOfElements] = l.m_elements[n];
	}
	m_numberOfElements += l.m_numberOfElements;
	return *this;
}

/////////////////////////////////////////////////////////////////////////////
//  Dereference an element in the list for lvalue purposes.
template < class O >
inline
O & List < O >::operator [](MType m) {
	CHECK(m < Count());
	return m_elements[m];
}

/////////////////////////////////////////////////////////////////////////////
//  Dereference an element in the list for rvalue purposes.
template < class O >
inline
const O & List < O >::operator [](MType m) const {
	CHECK(m < Count());
	return m_elements[m];
}

/////////////////////////////////////////////////////////////////////////////
//  How many elements are in the list.
template < class O >
inline
MType List < O >::Count(void) const {
	return m_numberOfElements;
}

/////////////////////////////////////////////////////////////////////////////
//  Empty out the list.  We may want to consider not resetting the malloc
// granularity parameter, or perhaps, halving it.
template < class O >
inline
void List < O >::Destroy(void) {
	SafeDelete(m_elements);
	m_numberOfElements = 0;
	m_nextAlloc = 1;
}

/////////////////////////////////////////////////////////////////////////////
//  Remove the last item from the list.
template < class O >
inline
void List < O >::Chop(void) {
	if (m_numberOfElements) {
		--m_numberOfElements;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Reverse the order of the items in the list.
template < class O >
inline
void List < O >::Flip(void) {
	MType m;

	for (m=0; m<m_numberOfElements / 2; ++m) {
		if (m != m_numberOfElements - 1 - m) {
			Swap(m_elements[m], m_elements[m_numberOfElements - 1 - m]);
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Allows us to work with maps safely and efficiently.
template < class O >
inline
MArray < O >::MArray(MType size, O * o) : m_size(size), m_elements(o) {
}

/////////////////////////////////////////////////////////////////////////////
//  Dereference one of these elements.
template < class O >
inline
O & MArray < O >::operator [](MType m) {
	CHECK(m < m_size);
	return m_elements[m];
}

/////////////////////////////////////////////////////////////////////////////
//  Dereference one of these elements, without modifying it.
template < class O >
inline
const O & MArray < O >::operator [](MType m) const {
	CHECK(m < m_size);
	return m_elements[m];
}

/////////////////////////////////////////////////////////////////////////////
//  How big am I.
template < class O >
inline
MType MArray < O >::Count(void) const {
	return m_size;
}

#if defined(TRACK_MEMORY)
	struct AllocRecord {
		AllocRecord * next;
		AllocRecord * prev;
		const char * f;
		int l;
		MType size;

		AllocRecord(const char * ff, int ll) : f(ff), l(ll) {}
	};

	struct MemBlock {
		int flags;
		AllocRecord * ar;
		MType count;
	};

	extern AllocRecord * g_allocRecords;
	extern MType g_maxMem;
	extern MType g_outstanding;

	/////////////////////////////////////////////////////////////////////////
	//  Same as above, except using the array format.
	template < class C >
	void SafeNew(C * & c, MType b, const char * f, int l) {
		AllocRecord * a = new AllocRecord(f, l);
		MemBlock * retval;
		C * dummy;

		if (g_allocRecords) {
			g_allocRecords->prev = a;
		}
		a->next = g_allocRecords;
		a->prev = NULL;
		g_allocRecords = a;
		a->size = b * sizeof(C);
		retval = (MemBlock *)malloc(sizeof(MemBlock) + b * sizeof(C));
		retval->ar = a;
		retval->count = b;
		retval->flags = 1;
		c = (C *)(retval + 1);
		g_outstanding += b * sizeof(C);
		if (g_outstanding > g_maxMem)
			g_maxMem = g_outstanding;
		while(b--) {
			dummy = new ((void *)(c + b)) C;
		}
	}

	/////////////////////////////////////////////////////////////////////////
	//  When we delete, we also remove the allocation record.  We also blank
	// out the pointer.
	template < class C >
	void SafeDelete(C * & c) {
		if (c) {
			MemBlock * r = (MemBlock *)c;
			AllocRecord * a;
			MType b;

			--r;
			a = r->ar;
			if (!r->flags) {
				Warn("Duplicate deallocation for block at %s, line %d.\n",
					 a->f,
					 a->l);
			}
			r->flags = 0;
			if (a->next) {
				a->next->prev = a->prev;
			}
			if (a->prev) {
				a->prev->next = a->next;
			} else {
				g_allocRecords = a->next;
			}
			delete a;

			b = r->count;
			g_outstanding -= b * sizeof(C);
			while (b--) {
				c[b].~C();
			}
			free((void *)r);
			c = NULL;
		}
	}
#else
	/////////////////////////////////////////////////////////////////////////
	//  A lot less work in this case.
	template < class C >
	inline
	void SafeNew(C * & c, MType b, const char * f, int l) {
		c = new C[b];
	}

	/////////////////////////////////////////////////////////////////////////
	//  Again, significantly less work.
	template < class C >
	inline
	void SafeDelete(C * & c) {
		delete [] c;
		c = NULL;
	}
#endif
