/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	The implementation of my thread manager.  For DPMI, preemptive multi-
threading is a real nuisance, so we have to do periodic polling to see if
a task switch is warranted.  In NT and Unix, this will not be necessary.  I
haven't yet figured out how I'm going to accomodate multi-processing in the
DOS case!  Hmmm. . .

****************************************************************************/

#include "all.h"

#define THREAD_GET_TYPE(x)				((x) >> THREAD_TYPE_SHIFT)
#define THREAD_GET_STATUS(x)			((x) & ((1 << THREAD_TYPE_SHIFT) - 1))
#define THREAD_MAKE_TYPE(x)				((x) << THREAD_TYPE_SHIFT)
#define THREAD_MASK						0x1f

enum {
	TIME_EXPIRED = 1,
	NO_MORE_TASKS
};

// 0 is an invalid frame number.  Don't let games last for more than 4 billion
// frames.
u32 g_currentFrame;
real g_currentFrameTime;
real g_currentFrameDeltaTime;
real g_desiredFrameTime = (1.0 / 20);
real g_processorSpeedInverse = (1.0 / 1000000000.0);

static Thread s_allThreads[THREAD_MAX_THREADS];
static ThreadJmpBuf s_initialPositions[THREAD_MAX_THREADS];
static ThreadJmpBuf s_retSpot;
static ThreadId s_id[THREAD_MAX_THREADS];
static ThreadId s_activeThread;
static u8 s_numberOfThreads;

extern "C" {
	extern void __cdecl ThreadPrepareThreadLaunches(Thread * threads,
													ThreadId * ids);
	extern void __cdecl ThreadSaveStackState(Thread & thread,
											 U8 * buffer);
	extern void __cdecl ThreadRestoreStackState(Thread & thread,
												U8 * buffer);
	extern int __cdecl ThreadSetJmp(ThreadJmpBuf * b);
	extern void __cdecl ThreadLongJmp(ThreadJmpBuf * b, int i);
};

/////////////////////////////////////////////////////////////////////////////
//  Open the thread manager.
static void ThreadOpen(void) {
	u32 u;

	for (u=0; u<THREAD_MAX_THREADS; ++u) {
		s_id[u] = u;
		s_allThreads[u].threadStatus = THREAD_NOT_A_THREAD;
	}
	ThreadPrepareThreadLaunches(s_allThreads, s_id);
	for (u=0; u<THREAD_MAX_THREADS; ++u) {
		memcpy(&s_initialPositions[u],
			   &s_allThreads[u].lastPosition,
			   sizeof(s_initialPositions[0]));
	}
	s_activeThread = (ThreadId)-1;
	s_numberOfThreads = 0;
}

static Initializer s_initializer(ThreadOpen);

/////////////////////////////////////////////////////////////////////////////
//  Okay, we've streamlined the system a bit; here's how it works:  when we
// want to launch a thread, we look for an available thread slot and if we
// find it, we insert the function call.  If no slots are available, we tell
// the user that the function call has failed.
u8 ThreadLaunchThread(ThreadEntryPoint entry,
					  void * entryArg,
					  ThreadId & id,
					  ThreadStatus & status,
					  u8 threadType,
					  MType stackSize) {
	u8 u;

	if (stackSize > THREAD_STACK_SIZE)
		return THREAD_MEM;
	else if (!stackSize)
		stackSize = THREAD_STACK_SIZE;

	for (u=0; u<THREAD_MAX_THREADS; ++u) {
		if (s_allThreads[u].threadStatus & THREAD_NOT_A_THREAD) {
			s_id[u] += THREAD_MASK + 1;
			id = s_id[u];

			s_allThreads[u].threadStatus = THREAD_MAKE_TYPE(threadType) |
										   THREAD_WAITING_TO_START;
			status = s_allThreads[u].threadStatus;
			s_allThreads[u].entryPoint = entry;
			s_allThreads[u].entryArg = entryArg;
			s_allThreads[u].netCPUTime = 0;
			s_allThreads[u].allotment = 0;
			s_allThreads[u].priority = THREAD_MIN_PRIORITY;
			++s_numberOfThreads;
			return THREAD_OK;
		}
	}
	return THREAD_MEM;
}

/////////////////////////////////////////////////////////////////////////////
//  Set a flag to suspend (or unsuspend) a thread.
u8 ThreadSuspend(ThreadId id, Bool suspend) {
	u8 u = id & (THREAD_MASK);
	u8 ty;

	if (id != s_id[u]) {
		return THREAD_INVALID;
	}
	Thread & t = s_allThreads[u];
	ty = THREAD_GET_TYPE(t.threadStatus);
	if (suspend) {
		t.threadStatus = THREAD_MAKE_TYPE(ty) | THREAD_SUSPENDED;
	} else {
		t.threadStatus = THREAD_MAKE_TYPE(ty) | THREAD_RUNNING;
	}
	return THREAD_OK;
}

/////////////////////////////////////////////////////////////////////////////
//  Make this a frame-cycled thread.
u8 ThreadMakeFrameCycled(ThreadId id) {
	u8 u = id & (THREAD_MASK);

	if (id != s_id[u]) {
		return THREAD_INVALID;
	}
	s_allThreads[u].threadStatus =
		THREAD_GET_STATUS(s_allThreads[u].threadStatus);
	s_allThreads[u].threadStatus |= THREAD_MAKE_TYPE(THREAD_FRAME_CYCLE);
	return THREAD_OK;
}

/////////////////////////////////////////////////////////////////////////////
//  Make this a critical thread.
u8 ThreadMakeCritical(ThreadId id) {
	u8 u = id & (THREAD_MASK);

	if (id != s_id[u]) {
		return THREAD_INVALID;
	}
	s_allThreads[u].threadStatus =
		THREAD_GET_STATUS(s_allThreads[u].threadStatus);
	s_allThreads[u].threadStatus |= THREAD_MAKE_TYPE(THREAD_CRITICAL);
	return THREAD_OK;
}

/////////////////////////////////////////////////////////////////////////////
//  Make this a timesliced thread.
u8 ThreadMakeTimesliced(ThreadId id) {
	u8 u = id & (THREAD_MASK);

	if (id != s_id[u]) {
		return THREAD_INVALID;
	}
	s_allThreads[u].threadStatus =
		THREAD_GET_STATUS(s_allThreads[u].threadStatus);
	s_allThreads[u].threadStatus |= THREAD_MAKE_TYPE(THREAD_TIMESLICED);
	s_allThreads[u].allotment = 0;
	return THREAD_OK;
}

/////////////////////////////////////////////////////////////////////////////
//  Make this a prioritized thread.
u8 ThreadMakePrioritized(ThreadId id) {
	u8 u = id & (THREAD_MASK);

	if (id != s_id[u]) {
		return THREAD_INVALID;
	}
	s_allThreads[u].threadStatus =
		THREAD_GET_STATUS(s_allThreads[u].threadStatus);
	s_allThreads[u].threadStatus |= THREAD_MAKE_TYPE(THREAD_PRIORITIZED);
	s_allThreads[u].priority = THREAD_MIN_PRIORITY;
	return THREAD_OK;
}

/////////////////////////////////////////////////////////////////////////////
//  Timesliced threads specify how much CPU time they need.  We aggregate
// that time and spend it when we can.
u8 ThreadAddTimeslice(ThreadId id,
					  real seconds) {
	u8 u = id & (THREAD_MASK);

	if (id != s_id[u]) {
		return THREAD_INVALID;
	}

	//	If we took extra time before, we dock it now.  If we quit early
	// before, we gain it back now.
	s_allThreads[u].allotment += seconds;
	return THREAD_OK;
}

/////////////////////////////////////////////////////////////////////////////
//  Raise my priority (if permissible).
u8 ThreadBoostPriority(ThreadId id) {
	u8 u = id & (THREAD_MASK);

	if (id != s_id[u]) {
		return THREAD_INVALID;
	}

	if (s_allThreads[u].priority < THREAD_MAX_PRIORITY)
		++s_allThreads[u].priority;
	return THREAD_OK;
}

/////////////////////////////////////////////////////////////////////////////
//  Make me less important.
u8 ThreadLowerPriority(ThreadId id) {
	u8 u = id & (THREAD_MASK);

	if (id != s_id[u]) {
		return THREAD_INVALID;
	}

	if (s_allThreads[u].priority > THREAD_MIN_PRIORITY)
		--s_allThreads[u].priority;
	return THREAD_OK;
}

/////////////////////////////////////////////////////////////////////////////
//  Tells the caller its thread Id.
u8 ThreadGetCurrentThread(ThreadId & id) {
	id = s_id[s_activeThread];
	return THREAD_OK;
}

/////////////////////////////////////////////////////////////////////////////
//  Return the status of this thread.
u8 ThreadGetThreadStatus(ThreadId id,
						 ThreadStatus & status) {
	u8 u = id & (THREAD_MASK);

	if (id != s_id[u]) {
		status = THREAD_NOT_A_THREAD;
		return THREAD_INVALID;
	} else {
		status = s_allThreads[u].threadStatus;
	}
	return THREAD_OK;
}

/////////////////////////////////////////////////////////////////////////////
//  How much time have we spent in this thread.
u8 ThreadGetCPUTime(ThreadId id, real & retval) {
	u8 u = id & (THREAD_MASK);

	if (id != s_id[u]) {
		return THREAD_INVALID;
	} else {
		retval = s_allThreads[u].netCPUTime;
		return THREAD_OK;
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Run the indicated thread.
static void ThreadRunThread(u8 u) {
	if (!ThreadSetJmp(&s_retSpot)) {
		real t = GetCPUSeconds();

		Thread & thread = s_allThreads[u];

		// Let's log the time for which we're exiting this
		// thread.
		if (s_activeThread != (ThreadId)-1) {
			Thread & oldThread = s_allThreads[s_activeThread & (THREAD_MASK)];
			oldThread.netCPUTime += t - oldThread.lastTime;
		}
		s_activeThread = u;
		s_allThreads[u].lastTime = t;
		if (THREAD_GET_STATUS(thread.threadStatus) == THREAD_WAITING_TO_START) {
			u8 ty = THREAD_GET_TYPE(thread.threadStatus);

			thread.threadStatus = THREAD_MAKE_TYPE(ty) | THREAD_RUNNING;
		}
	
		ThreadLongJmp(&thread.lastPosition, 1);
	}
}
	
extern "C" {
	/////////////////////////////////////////////////////////////////////////////
	//  Sometimes, we need to kill a thread.  This function handles that for us.
	u8 __cdecl ThreadKillThread(ThreadId id) {
		u8 u = id & (THREAD_MASK);
	
		if ((id != s_id[u]) ||
			(THREAD_GET_STATUS(s_allThreads[u].threadStatus) == THREAD_NOT_A_THREAD)) {
			return THREAD_INVALID;
		}
		s_allThreads[u].threadStatus = THREAD_NOT_A_THREAD;
		memcpy(&s_allThreads[u].lastPosition,
			   &s_initialPositions[u],
			   sizeof(s_initialPositions[0]));
		--s_numberOfThreads;
		if (u == s_activeThread)
			ThreadLongJmp(&s_retSpot, 1);
		return THREAD_OK;
	}
};

/////////////////////////////////////////////////////////////////////////////
//  Kill all threads.
void ThreadKillAll(void) {
	u8 v;
	u8 u = s_activeThread;

	for (v=0; v<THREAD_MAX_THREADS; ++v) {
		if (u == v) {
			continue;
		}
		ThreadKillThread(s_id[v]);
	}
	ThreadKillThread(s_id[u]);
}

/////////////////////////////////////////////////////////////////////////////
//  Kill the zombie indicated by the ID passed in.  We ignore this call.
u8 ThreadKillZombie(ThreadId id) {
	return THREAD_OK;
}

/////////////////////////////////////////////////////////////////////////////
//  Kill both the thread and the zombie.  This function chains to the regular
// kill, since our zombies are implicit.
u8 ThreadKillThreadAndZombie(ThreadId id) {
	return ThreadKillThread(id);
}

/////////////////////////////////////////////////////////////////////////////
//  To suspend a thread, we copy its stack into a caller-allocated buffer
// and kill the thread.  HERE I suspect trouble if a thread tries to
// suspend itself.  In particular, I expect it will come back here when
// restored!  There may be a way to prevent that - for instance, we may
// be able to kill the thread first, then save the stack state.
u8 ThreadSuspendThread(ThreadId id, ThreadSaveBuffer & buffer) {
	u8 u = id & (THREAD_MASK);

	if ((id != s_id[u]) ||
		(s_allThreads[u].threadStatus & THREAD_NOT_A_THREAD)) {
		return THREAD_INVALID;
	}
	buffer.CopyThreadInfo(s_allThreads[u]);
	ThreadSaveStackState(s_allThreads[u], buffer.GetBuffer());
	return ThreadKillThreadAndZombie(id);
}

/////////////////////////////////////////////////////////////////////////////
//  To resume a suspended thread, we get the next available slot, copy the
// stack frame back into the stack, and restore all the information about
// the thread.
u8 ThreadResumeThread(ThreadId & id, ThreadSaveBuffer & buffer) {
	u8 u;

	for (u=0; u<THREAD_MAX_THREADS; ++u) {
		if (s_allThreads[u].threadStatus & THREAD_NOT_A_THREAD) {
			s_id[u] += THREAD_MASK + 1;
			id = s_id[u];

			buffer.RestoreThreadInfo(s_allThreads[u]);
			ThreadRestoreStackState(s_allThreads[u], buffer.GetBuffer());
			++s_numberOfThreads;
			return THREAD_OK;
		}
	}
	return THREAD_MEM;
}

/////////////////////////////////////////////////////////////////////////////
//  We don't have preemptive multitasking, so we have to do some polling in
// the cooperative case.  In the preemptive case, these functions aren't
// called.
//  The parameter here indicates whether we are yielding (TRUE) or polling
// (FALSE).
void ThreadYield(void) {
	Thread & t = s_allThreads[s_activeThread];
	u8 u = THREAD_GET_TYPE(t.threadStatus);
	real r;

	switch(u) {
		case THREAD_CRITICAL:
			return;
		case THREAD_TIMESLICED:
		case THREAD_PRIORITIZED:
			r = GetCPUSeconds();
			r -= t.lastTime;
			if ((t.allotment -= r) > (real)0.0) {
				return;
			}
		case THREAD_FRAME_CYCLE:
			if (!ThreadSetJmp(&t.lastPosition)) {
				ThreadLongJmp(&s_retSpot, 1);
			}
	}
}

/////////////////////////////////////////////////////////////////////////
//  Again, because we don't have pre-emptive multitasking, we need this
// function to enter our pm emulation mode.
//  06Aug99, 23:17:  Okay, gang, thread logic is being substantially
// reworked.  Basically, when we enter multithread mode, we start a
// routine that cycles through one frame at a time, executing the threads
// according to the thread execution rules.
void ThreadEnterMultiThreadMode(void) {
	u8 u;
	real sofar;

	g_currentFrame = 0;
	g_currentFrameTime = 0;
	g_currentFrameDeltaTime = 0;

	while (s_numberOfThreads) {
		real lastTime = g_currentFrameTime;
		u8 totalPriority = 0;
		u8 sliced[THREAD_MAX_THREADS];
		u8 nSliced = 0;
		u8 prioritized[THREAD_MAX_THREADS];
		u8 nPrioritized=0;

		g_currentFrameTime = GetCPUSeconds();
		g_currentFrameDeltaTime = g_currentFrameTime - lastTime;
		++g_currentFrame;

		// Okay, here's where we'll be doing all of the per-frame thread
		// logic.
		for (u=0; u<THREAD_MAX_THREADS; ++u) {
			if (THREAD_GET_STATUS(s_allThreads[u].threadStatus) == THREAD_NOT_A_THREAD)
				continue;
			if (THREAD_GET_STATUS(s_allThreads[u].threadStatus) == THREAD_SUSPENDED)
				continue;

			u8 v = THREAD_GET_TYPE(s_allThreads[u].threadStatus);

			switch(v) {
				case THREAD_FRAME_CYCLE:
					ThreadRunThread(u);
					break;
				case THREAD_CRITICAL:
					ThreadRunThread(u);
					break;
				case THREAD_PRIORITIZED:
					prioritized[nPrioritized++] = u;
					totalPriority += s_allThreads[u].priority;
					break;
				case THREAD_TIMESLICED:
					sliced[nSliced++] = u;
					break;
			}
		}
		// Okay, our per-frame threads executed until they yielded, and
		// our critical threads executed 'til completion.  Now, let's
		// resolve the timeslice / priority issue.
		sofar = GetCPUSeconds() - g_currentFrameTime;
		sofar = g_desiredFrameTime - sofar;
		if (sofar > (real)0.0) {
			if (totalPriority) {
				real dt;
				real denom = Reciprocoal((real)totalPriority);
		
				dt = sofar;
	
				dt *= (real)totalPriority;
				dt *= (real)(1.0 / (THREAD_MAX_PRIORITY * THREAD_MAX_THREADS + 1));
				for (u=0; u<nPrioritized; ++u) {
					u8 v = prioritized[u];
	
					s_allThreads[v].allotment = dt *
												(real)(s_allThreads[v].priority) *
												denom;
					ThreadRunThread(v);
				}
			}
			sofar = GetCPUSeconds() - g_currentFrameTime;
			sofar = g_desiredFrameTime - sofar;
			if ((nSliced) && (sofar > (real)0.0)) {
				// Sort the timesliced threads.
				for (u=0; u<nSliced-1; ++u) {
					u8 v;
		
					for (v=u+1; v<nSliced; ++v) {
						if (s_allThreads[sliced[u]].allotment <
							s_allThreads[sliced[v]].allotment) {
							Swap(sliced[u], sliced[v]);
						}
					}
				}
		
				// Now, we go through all of the timesliced threads and execute
				// until we've run out of time in our frame.
				for (u=0; u<nSliced - 1; ++u) {
					u8 v = sliced[u];

					ThreadRunThread(v);
					sofar = GetCPUSeconds() - g_currentFrameTime;
					sofar = g_desiredFrameTime - sofar;
					if (sofar < (real)0.0)
						break;
				}
			}
		}
	}
}
