/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	These are View functions that have a lot in common from one platform to
the next.

****************************************************************************/

#include "all.h"

struct RGB {
	U8 blue;
	U8 green;
	U8 red;
};

struct RGBQuad {
	U8 blue;
	U8 green;
	U8 red;
	U8 reserved;
};

struct BMPHead {
	S16 type;
	U32 size;
	U32 reserved;
	U32 offBits;
};

struct BMPInfo {
	U32 size;
	U32 width;
	U32 height;
	S16	planes;
	S16	bitCount;
	U32 compression;
	U32 sizeImage;
	U32 xPelsPerMeter;
	U32 yPelsPerMeter;
	U32 clrUsed;
	U32 clrImportant;
};

/////////////////////////////////////////////////////////////////////////////
//  This function will import a BMP file to a map.
void View::ImportFile(const FString & fname,
					  Map < Color > & m) const {
	BMPHead bmpHead;
	BMPInfo bmpInfo;
	RGBQuad * map;
	char name[256];
	FILE * file;
	u16 width, height;
	u16 colors;
	u16 i, j;

	DirFormFilename(name, fname);

	if (!(file = fopen(name, "rb"))) {
		Warn("Unable to open file %s.\n", name);
		return;
	}

	fread(&bmpHead, sizeof(bmpHead), 1, file);
	fread(&bmpInfo, sizeof(bmpInfo), 1, file);

	// Sanity checks.
	if (bmpHead.type != 0x4d42) {
		Warn("Invalid BMP file containing (bmpHead.type fails)\n");
		return;
	}
	if (bmpInfo.size != sizeof(bmpInfo)) {
		Warn("Invalid BMP info header size for file %s.\n",
			 name);
		return;
	}
	if ((bmpInfo.bitCount != 8) &&
		(bmpInfo.bitCount != 24) &&
		(bmpInfo.bitCount != 4)) {
		Warn("Unsupported bit count in bmp file %s.\n",
			 name);
		return;
	}
	if (bmpInfo.compression) {
		Warn("Compressed BMP files not supported (file %s).\n",
			 name);
		return;
	}

	// Allocate the map.
	width = bmpInfo.width;
	height = bmpInfo.height;
	m.Allocate(width, height, __FILE__, __LINE__);
	m.Open(m.WRITE);
	SafeNew(map, width * height, __FILE__, __LINE__);
	if (bmpInfo.bitCount == 8) {
		RGBQuad rgbQuad[256];
		unsigned char * tbuf;

		if (bmpInfo.clrUsed) {
			colors = bmpInfo.clrUsed;
		} else {
			colors = 1 << bmpInfo.bitCount;
		}
		fread(rgbQuad, sizeof(RGBQuad), colors, file);

		SafeNew(tbuf, width * height, __FILE__, __LINE__);
		fread(tbuf, 1, width * height, file);
		for (j=0; j<height; ++j) {
			for (i=0; i<width; ++i) {
				map[j * width + i].red = rgbQuad[tbuf[j * width + i]].red;
				map[j * width + i].green = rgbQuad[tbuf[j * width + i]].green;
				map[j * width + i].blue = rgbQuad[tbuf[j * width + i]].blue;
			}
		}
		SafeDelete(tbuf);
	} else if (bmpInfo.bitCount == 4) {
		RGBQuad rgbQuad[16];
		unsigned char * tbuf;

		if (bmpInfo.clrUsed) {
			colors = bmpInfo.clrUsed;
		} else {
			colors = 1 << bmpInfo.bitCount;
		}
		fread(rgbQuad, sizeof(RGBQuad), colors, file);

		SafeNew(tbuf, width * height / 2, __FILE__, __LINE__);
		fread(tbuf, 1, width * height / 2, file);
		for (j=0; j<height; ++j) {
			for (i=0; i<width; ++i) {
				u16 index;

				if (i & 1) {
					index = tbuf[(j * width + i) / 2] & 0xf;
				} else {
					index = tbuf[(j * width + i) / 2] >> 4;
				}
				map[j * width + i].red = rgbQuad[index].red;
				map[j * width + i].green = rgbQuad[index].green;
				map[j * width + i].blue = rgbQuad[index].blue;
			}
		}
		SafeDelete(tbuf);
	} else {
		RGB * rgb;

		SafeNew(rgb, width * height, __FILE__, __LINE__);
		fread(rgb, sizeof(RGB), width * height, file);
		for (i=0; i<width * height; ++i) {
			map[i].red = rgb[i].red;
			map[i].green = rgb[i].green;
			map[i].blue = rgb[i].blue;
		}
		SafeDelete(rgb);
	}

	// Flip the image.
	for (j=0; j<height / 2; ++j) {
		u16 k = height - j - 1;

		for (i=0; i<width; ++i) {
			unsigned char c;
			RGBQuad * r1 = &map[j * width + i];
			RGBQuad * r2 = &map[k * width + i];

			c = r1->red;
			r1->red = r2->red;
			r2->red = c;

			c = r1->green;
			r1->green = r2->green;
			r2->green = c;

			c = r1->blue;
			r1->blue = r2->blue;
			r2->blue = c;
		}
	}
	// Conversion time!
	for (j=0; j<height; ++j) {
		for (i=0; i<width; ++i) {
			RGBQuad & q = map[j * width + i];

			m[j][i] = MapColor(q.red * (real)(1.0 / 255.0),
							   q.green * (real)(1.0 / 255.0),
							   q.blue * (real)(1.0 / 255.0));
		}
	}
	m.Close();
	SafeDelete(map);
}

/////////////////////////////////////////////////////////////////////////////
//  This function will export a map to a BMP file.
void View::ExportFile(const FString & fname,
					  Map < Color > & m) const {
	BMPHead bmpHead;
	BMPInfo bmpInfo;
	char name[256];
	FILE * file;
	u16 width, height;
	u16 i, j;

	DirFormFilename(name, fname);

	if (!(file = fopen(name, "wb"))) {
		Warn("Unable to open file %s.\n", name);
		return;
	}

	bmpHead.type = 0x4d42;
	bmpHead.size = sizeof(bmpHead) +
				   sizeof(bmpInfo) +
				   m.GetWidth() * m.GetHeight() * sizeof(RGB);
	bmpHead.reserved = 0;
	bmpHead.offBits = sizeof(bmpHead) + sizeof(bmpInfo);

	bmpInfo.size = sizeof(bmpInfo);
	bmpInfo.bitCount = 24;
	bmpInfo.compression = 0;
	bmpInfo.width = m.GetWidth();
	bmpInfo.height = m.GetHeight();
	bmpInfo.planes = 1;
	bmpInfo.sizeImage = 0;
	bmpInfo.xPelsPerMeter = 0;
	bmpInfo.yPelsPerMeter = 0;
	bmpInfo.clrUsed = 0;
	bmpInfo.clrImportant = 0;

	fwrite(&bmpHead, sizeof(bmpHead), 1, file);
	fwrite(&bmpInfo, sizeof(bmpInfo), 1, file);

	width = bmpInfo.width;
	height = bmpInfo.height;
	m.Open(m.READ);

	// Okay, we have to flip the image, so let's go bottom-up.
	for (j=height - 1; j >= 0; --j) {
		for (i=0; i<width; ++i) {
			RGB rgb;
			real r, g, b;

			MapColor(m[j][i], r, g, b);
			rgb.red		= r * 255.9999;
			rgb.green	= g * 255.9999;
			rgb.blue	= b * 255.9999;
			fwrite(&rgb, sizeof(rgb), 1, file);
		}
		if (!j)
			break;
	}
	m.Close();
	fclose(file);
}

