/****************************************************************************

	Written by John Marco Panettiere for John Marco Panettiere.
	Copyright 1998, 1999, 2000 by John Marco Panettiere.
	All rights reserved.

	Windows 2001!  My own windowing system.

****************************************************************************/

#ifndef WINDOW_H
	#define WINDOW_H

	enum {
		UNSELECTED		= 0,
		SELECTED,
		BORDER,
		BACKGROUND,
		TEXT,

		NUMBER_OF_TEXT_COLORS
	};

	extern color g_colors[];

	class Window;

	/////////////////////////////////////////////////////////////////////////
	//	 A GUIElement is a potential recipient of a mouse click or a drag
	// or whatever.  The graphics of a GUIElement are managed separately.
	class GUIElement : public ActiveGraphic {
		public:
			GUIElement(void);
			virtual Bool Click(u16 x, u16 y) = 0;
			virtual void Render(void) = 0;

			virtual u16 GetWidth(void) const =0;
			virtual u16 GetHeight(void) const =0;

			void Hide(void);
			void Show(void);
			void SetX(u16 x);
			void SetY(u16 y);

			Bool IsDirty(void);
		protected:
			Bool m_visible;
			u16 m_x, m_y;
		private:
			u8 m_needToUpdate;
	};

	/////////////////////////////////////////////////////////////////////////
	//	A checkbox is a item that is either selected or deselected.  We'll
	// keep width and height constant for simplicity's sake.
	class GUICheckBox : public GUIElement {
		public:
			GUICheckBox(void);

			virtual u16 GetWidth(void) const;
			virtual u16 GetHeight(void) const;

			void SetSpacing(u8 spaces);
			void SetTitle(const FString & title);
			void SetDims(u16 x, u16 y);
			virtual Bool Click(u16 x, u16 y);
			virtual void Render(void);
		protected:
			virtual Bool QuerySelect(void) = 0;
			virtual void Select(void) = 0;
			virtual void Deselect(void) = 0;

			IString m_title;
			u8 m_spacing;
		private:
	};

	/////////////////////////////////////////////////////////////////////////
	//	A radio checkbox is a set of mutually exclusive checkboxes.
	class GUIRadioCheckBox : public GUIElement {
		public:
			GUIRadioCheckBox(void);

			virtual u16 GetWidth(void) const;
			virtual u16 GetHeight(void) const;

			void SetTitle(const FString & title);
			void SetCaption(u8 item, const FString & caption);
			void SetDims(u16 x, u16 y, u8 count);
			virtual Bool Click(u16 x, u16 y);
			virtual void Render(void);
		protected:
			virtual u8 QuerySelect(void) = 0;
			virtual void Select(u8 item) = 0;
			virtual void Deselect(u8 item) = 0;

			IString m_title;
			Array < IString > m_captions;
			u8 m_selection;
		private:
	};

	/////////////////////////////////////////////////////////////////////////
	//	Windows have several different panels.  Each panel has its own rule
	// for redrawing the client area.
	class Panel : public ActiveGraphic {
		public:
			void SetParent(Window * parent);

			virtual void Click(u16 x, u16 y);
			virtual void RClick(u16 x, u16 y);

			virtual void SetDims(u16 x, u16 y, u16 w, u16 h);
			virtual void Redraw(void);
			virtual void CursorUp(void)=0;
			virtual void CursorDown(void)=0;
			virtual void PageUp(void)=0;
			virtual void PageDown(void)=0;

			virtual void MakeDirty(void);

			virtual void GenerateContent(void)=0;

			void AddGUIElement(GUIElement * g);
		protected:
			u8 m_needToUpdate;
			Window * m_parent;
			Rect m_rect;
			List < GUIElement * > m_guiElements;
		private:
	};

	/////////////////////////////////////////////////////////////////////////
	//	This type of panel fills the client area with a body of text.
	class TextPanel : public Panel {
		public:
			TextPanel(void);

			virtual void Click(u16 x, u16 y);

			virtual void SetDims(u16 x, u16 y, u16 w, u16 h);
			virtual void Redraw(void);
			virtual void CursorUp(void);
			virtual void CursorDown(void);
			virtual void PageUp(void);
			virtual void PageDown(void);

			virtual void Goto(u32 line);

			virtual void GenerateContent(void)=0;
		protected:
			Array < FString > m_content;
			u8 m_needToUpdate;
		private:
	};

	/////////////////////////////////////////////////////////////////////////
	//	A dummy text panel class for the dummy text generator.
	class DummyPanel : public TextPanel {
		public:
			virtual void GenerateContent(void);
		protected:
		private:
	};

	/////////////////////////////////////////////////////////////////////////
	//	This is for panels that contain a finite list of strings that do not
	// need to be procedurally generated.
	class IListTextPanel : public TextPanel {
		public:
			virtual void GenerateContent(void);

			virtual void CursorUp(void);
			virtual void CursorDown(void);
			virtual void PageUp(void);
			virtual void PageDown(void);

			virtual void Goto(u32 line);

			IString & operator [](MType m);
			u32 Bottom(void);
		protected:
			List < IString > m_data;
			MType m_lineNumber;
		private:
	};

	/////////////////////////////////////////////////////////////////////////
	//	Text panels that contain a fixed number of strings that do not need
	// to be procedurally generated.
	class FListTextPanel : public TextPanel {
		public:
			void InsertString(const FString & f);
			virtual void GenerateContent(void);
		protected:
		private:
	};

	// Window mouse hotspots have their own augmentations.
	class WindowSpot : public MouseHotspot {
		public:
			void SetArguments(Window * w, u8 tab);
			virtual void Click(u16 x, u16 y);
		protected:
		private:
			Window * m_window;
			U8 m_tab;
	};
	class ClientSpot : public MouseHotspot {
		public:
			virtual ~ClientSpot(void);

			void SetArguments(Window * w);
			virtual void Click(u16 x, u16 y);
			virtual void RClick(u16 x, u16 y);
		protected:
		private:
			Window * m_window;
	};
	struct WindowPanel {
		FString name;
		Panel * redraw;
	};

	/////////////////////////////////////////////////////////////////////////
	//	This is what all multi-tabbed windows have in common.
	class Window : public ActiveGraphic {
		public:
			Window(void);
			~Window(void);

			void SetXYWH(u16 x, u16 y, u16 w, u16 h);
			void AddPanel(const FString & name,
						  Panel * redraw);
			void GetPanelDims(u16 & x, u16 & y, u16 & w, u16 & h);

			// When a change is effected to the window, we totally update
			// the client on both pages.  Once this is done, no further
			// updates need to occur.
			void Refresh(void);
			void ForceRedraw(void);
			virtual void SetTab(u8 tab);
			u8 GetTab(void);

			void Activate(void);
			void Deactivate(void);

			void CursorUp(void);
			void CursorDown(void);
			void PageUp(void);
			void PageDown(void);

			static void UpdateAll(void);
		protected:
			friend class WindowSpot;
			friend class ClientSpot;

			List < WindowPanel > m_panels;
			u16 m_x, m_y, m_w, m_h;
			u8 m_needToUpdate;
			MType m_selection;
			MouseHotspotGroup m_tabs;
			static List < Window * > s_allWindows;
		private:
	};

	extern FontHandle g_windowFont;

	#include "window.ipp"

#endif
